/**
 * @file uAC.c
 * @author Thomas Weinhold
 * @date 14.02.2011
 * @version 0.1
 */

#include <uAC.h>
#include <stdarg.h>

//! The Status vars.
struct {
	unsigned char PromtIsWritten:1;
	unsigned char ReadyToResolveLine:1;
	unsigned char ReadyToExecute:1;
	unsigned char EscapeSequence:2;
} _status;

//!The uAC input buffer.
struct {
	unsigned int read;
	unsigned int write;
	unsigned int minfree;
	char buffer[_uAC_InBufferSize];
} _inbuff;

//!The uAC output buffer.
struct {
	unsigned int read;
	unsigned int write;
	unsigned int minfree;
	char buffer[_uAC_OutBufferSize];
} _outbuff;

//!Stores user input.
struct {
	unsigned int pointer;
	char buffer[_uAC_LineLength+1];
} _history[_uAC_HistorySize+1];

//!Stores the attached commands
struct {
	char *name;
	void (*function)(int,char*[]);
} _cmdlib[uAC_CmdLibLength];

//!Indicates which history point is selected
unsigned int _history_pointer=0;

//!Put a char into the outbuffer
void _putchar ( char);

//!Put a string to the outbuffer
inline void _putstring (char *);

//!Put a uint to the outbuffer.
void _putuint (unsigned int);

//!Write the newline string.
inline void _newline (void);

//!Clear the prompt line.
void _prompt_clear (void);

//!Write the prompt line.
void _prompt_write (void);

//!Process an incoming char.
int _process_char (void);

//!Read the command line and execute the command.
void _resolve_line (void);

//!Push the actual command line into the history.
void _push_history (void);

//!The help command.
void _print_help (void);

void uAC_Init (void) {
	unsigned int i=0;
	_inbuff.read=0;
	_inbuff.write=0;
	_inbuff.minfree=(unsigned int) -1;
	_outbuff.read=0;
	_outbuff.write=0;
	_outbuff.minfree=(unsigned int) -1;

	for (i=0;i<=_uAC_HistorySize;i++) _history[i].pointer=0;
	_history_pointer=0;

	for (i=0;i<uAC_CmdLibLength;i++) _cmdlib[i].name=0;

	_newline();
	_prompt_write();

	_status.ReadyToResolveLine=0;
	_status.ReadyToExecute=0;
	_status.EscapeSequence=0;
}

void uAC_Task (void) {
	if (_status.ReadyToResolveLine){
		_resolve_line(); //!execute command
	}
	else if (_inbuff.read != _inbuff.write) {
		_process_char(); //!read input buffer
	}
}

int uAC_RxInput (char Rx) {
	unsigned int NextWrite = ( _inbuff.write + 1 ) & (_uAC_InBufferSize-1);
	unsigned int MinFree=0;
	int Ret=0;

	if(NextWrite != _inbuff.read) {
		_inbuff.write = NextWrite;
		_inbuff.buffer[NextWrite] = Rx;
		MinFree=(_inbuff.read - _inbuff.write - 1) & (_uAC_InBufferSize-1);
		if (MinFree < _inbuff.minfree) _inbuff.minfree=MinFree;
		Ret=1;
	}

	return Ret;
}

int uAC_TxAvailable (void) {
	return ( (_outbuff.read == _outbuff.write) ? 0:1);
}

char uAC_GetTxChar (void) {
	if (uAC_TxAvailable()) {
		return _outbuff.buffer[_outbuff.read = (_outbuff.read + 1) & (_uAC_OutBufferSize-1)];
	}

	return 0;
}

int uAC_Attach(char *name, void (*func)(int,char*[])) {
	int Ret=0;
	unsigned int i=0;

	for (i=0;i<uAC_CmdLibLength;i++) {
		if (_cmdlib[i].name==0) {
			_cmdlib[i].name=name;
			_cmdlib[i].function=func;
			Ret=1;
			break;
		}
	}

	return Ret;
}

void uAC_Print(char * Format, ...) {
	unsigned int i=0;
	char *c_ptr,c,RestorePrompt=_status.PromtIsWritten;
	int s_i;
	unsigned int u_i;

	if (RestorePrompt) _prompt_clear();

	va_list ap;
	va_start(ap,Format);

	for (i=0; Format[i]!='\0'; i++) {
		if (Format[i]=='%') {
			i++;
			switch (Format[i]) {
			case '%':
				_putchar('%');
				break;
			case 's':
				c_ptr = va_arg(ap,char *);
				_putstring(c_ptr);
				break;
			case 'c':
				c = (char) va_arg(ap,int);
				_putchar(c);
				break;
			case 'i':
				s_i = va_arg(ap,int);
				if (s_i < 0) {
					_putchar('-');
					s_i*=-1;
				}
				_putuint((unsigned int) s_i);
				break;
			case 'u':
				u_i = va_arg(ap,unsigned int);
				_putuint(u_i);
				break;
			case '\0':
				i--;
			default:
				_putchar('%');
				_putchar(Format[i]);
			}
		}
		else if (Format[i]=='\n') {
			_newline();
		}
		else {
			_putchar(Format[i]);
		}
	}
	va_end(ap);
	if (RestorePrompt) _prompt_write();
}

int uAC_StringCompare (char *a, char *b)
{
	unsigned int i;
	int Ret=1;

	for (i=0;a[i]!='\0' && b[i]!='\0';i++) {
		if (a[i]!=b[i]) Ret=0;
	}
	if (a[i]!=b[i]) Ret=0;

	return Ret;
}

void _putchar ( char C) {
	unsigned int NextWrite = ( _outbuff.write + 1 ) & (_uAC_OutBufferSize-1);
	unsigned int MinFree=0;

	if(NextWrite != _outbuff.read) {
		_outbuff.write = NextWrite;
		_outbuff.buffer[NextWrite] = C;
		MinFree=(_outbuff.read - _outbuff.write - 1) & (_uAC_OutBufferSize-1);
		if (MinFree < _outbuff.minfree) _outbuff.minfree=MinFree;
	}
}

inline void _putstring (char *s) {
	for (;*s!='\0';) _putchar(*(s++));
}

void _putuint (unsigned int ux) {
	if (ux==0) {
		_putchar('0');
		return;
	}
	char c_arr[11], mod=0;
	int i=0;
	for (i=0; ux; i++) {
		mod=(char) (ux%10);
		ux=(ux-((unsigned int)mod))/10;
		c_arr[i] = mod + '0';
	}
	for (;i>=0;i--) _putchar(c_arr[i]);
}

inline void _newline (void) {
	_putstring(_uAC_NewLineString);
}

void _prompt_clear (void) {
	int i;
	for (i=0;_uAC_Prompt[i]!='\0';i++) _putstring("\b \b");
	for (i=0; i < _history[0].pointer; i++) _putstring("\b \b");
	_status.PromtIsWritten=0;
}

void _prompt_write (void) {
	unsigned int i=0;
	_putstring(_uAC_Prompt);
	for (i=0; i < _history[0].pointer; i++) _putchar(_history[0].buffer[i]);
	_status.PromtIsWritten=1;
}

void _push_history (void) {
	if (!_uAC_HistorySize) return;

	unsigned int i=0,i2=0;
	unsigned char CmdIsEqual=1;

	//!Check if actual command line equals the last one in the history.
	if (_history[0].pointer != _history[1].pointer) {
		CmdIsEqual=0;
	}
	else {
		for (i=0;i<_history[0].pointer;i++) {
			if (_history[0].buffer[i]!=_history[1].buffer[i]) {
				CmdIsEqual=0;
				break;
			}
		}
	}

	//Push the actual line into the History
	if (!CmdIsEqual) {
		for (i=_uAC_HistorySize;i!=0;i--) {
			for (i2=0;i2<_history[i-1].pointer;i2++) {
				_history[i].buffer[i2]=_history[i-1].buffer[i2];
			}
			_history[i].pointer=_history[i-1].pointer;
		}
	}
}

void _print_help (void) {
	unsigned int i;
	for (i=0;i<uAC_CmdLibLength;i++) {
		if(_cmdlib[i].name==0) break;
		uAC_Print("%s\n",_cmdlib[i].name);
	}
	uAC_Print("\n{free buffer: in %u / out %u}\n",_inbuff.minfree,_outbuff.minfree);
}

int _process_char(void)
{
	char pc=0; //!process char
	unsigned int i; //!Loop int

	//! Get a char from the input buffer.
	if (_inbuff.read != _inbuff.write) {
		pc = _inbuff.buffer[_inbuff.read = (_inbuff.read + 1) & (_uAC_InBufferSize-1)];
	}
	else {
		return 0;
	}

	//! Process escape sequences.
	if (_status.EscapeSequence==1) {
		if (pc=='[') {
			_status.EscapeSequence=2;
		}
		else {
			if (_history[0].pointer!=_uAC_LineLength) {
				_history[0].buffer[_history[0].pointer++]=pc;
				_putchar(pc);
			}
			_status.EscapeSequence=0;
		}
		return 0;
	}
	else if (_status.EscapeSequence==2) {
		if (pc=='A') {
			_status.EscapeSequence=0;
			if (!_uAC_HistorySize) return 0;
			if (_history_pointer<_uAC_HistorySize) {
				_prompt_clear();
				_history_pointer++;
				for (i=0;i<_history[_history_pointer].pointer;i++)
					_history[0].buffer[i]=_history[_history_pointer].buffer[i];
				_history[0].pointer=_history[_history_pointer].pointer;
				_prompt_write();
			}
		}
		else if (pc=='B') {
			_status.EscapeSequence=0;
			if (!_uAC_HistorySize) return 0;
			if (_history_pointer>1) {
				_prompt_clear();
				_history_pointer--;
				for (i=0;i<_history[_history_pointer].pointer;i++)
					_history[0].buffer[i]=_history[_history_pointer].buffer[i];
				_history[0].pointer=_history[_history_pointer].pointer;
				_prompt_write();
			}
			else if (_history_pointer==1) {
				_prompt_clear();
				_history_pointer=0;
				_history[0].pointer=0;
				_prompt_write();
			}
		}
		else {
			_status.EscapeSequence=0;
		}
		return 0;
	}

	//! Process other chars.
	switch (pc) {
	case '\t':
		break;
	case 27:
		_status.EscapeSequence=1;
		break;
	case '\r':
	case '\n':
		_newline();
		_status.PromtIsWritten=0;
		_push_history();
		_history_pointer=0;
		_status.ReadyToResolveLine=1;
		break;
	case '\b':
		if (_history[0].pointer!=0) {
			_history[0].pointer--;
			_putstring("\b \b");
		}
		break;
	default:
		if (_history[0].pointer!=_uAC_LineLength) {
			_history[0].buffer[_history[0].pointer++]=pc;
			_putchar(pc);
		}
	}

	return (1);
}

void _resolve_line (void) {
	//!terminate command line
	_history[0].buffer[_history[0].pointer]='\0';

	//!Reset Prompt
	unsigned int i=0,i2=0,argc=0,arg_begins=0;

	//!Count arguments
	for (i=0;i < _history[0].pointer;i++) {
		if (_history[0].buffer[i]==' ' || _history[0].buffer[i]=='\t') {
			arg_begins=0;
		}
		else {
			if (!arg_begins) {
				arg_begins=1;
				argc++;
			}
		}
	}

	//Exit if there are no arguments
	if (!argc) {
		_prompt_write();
		_status.ReadyToResolveLine=0;
		return;
	}

	//!The arguments vector.
	char *argv[argc];

	//!Get pointers
	i2=0;
	arg_begins=0;
	for (i=0;i < _history[0].pointer;i++) {
		if (_history[0].buffer[i]==' ' || _history[0].buffer[i]=='\t') {
			_history[0].buffer[i]='\0';
			arg_begins=0;
		}
		else {
			if (!arg_begins) {
				arg_begins=1;
				argv[i2++]=&_history[0].buffer[i];
			}
		}
	}

	//!search the cmd in the cmdlib
	if (uAC_StringCompare("help",argv[0])) {
		_print_help();
	}
	else {
		for (i=0;i<uAC_CmdLibLength;i++) {
			if (uAC_StringCompare(argv[0],_cmdlib[i].name)) break;
		}
		if (i==uAC_CmdLibLength) {
			uAC_Print("command not found!\n");
		}
		else {
			_cmdlib[i].function(argc-1,&argv[1]);
		}
	}

	//!replace '\0' with ' '
	for (i=0;i < _history[0].pointer;i++) {
		if (_history[0].buffer[i]=='\0') _history[0].buffer[i]=' ';
	}

	_newline();
	_history[0].pointer=0;
	_prompt_write();
	_status.ReadyToResolveLine=0;
}
